"""Release [0.34.0].

Revision ID: 0.34.0
Revises: 03742aa7fdd7
Create Date: 2023-02-20 09:15:55.777994

"""


# revision identifiers, used by Alembic.
revision = "0.34.0"
down_revision = "03742aa7fdd7"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
