"""Release [0.40.1].

Revision ID: 0.40.1
Revises: 0.40.0
Create Date: 2023-05-26 18:08:22.849722

"""


# revision identifiers, used by Alembic.
revision = "0.40.1"
down_revision = "0.40.0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
